# Example
# ~~~~~~~
# from aiko_services.main.proxy import ProxyAllMethods, proxy_trace, Example
# example_proxy = ProxyAllMethods("Example", Example("v0"), proxy_trace)
# example_proxy.function_0("value_0", argument_1="value_1")
#
# Resources
# ~~~~~~~~~
# Python wrapt package
# - https://wrapt.readthedocs.io/en/latest/wrappers.html
#
# To Do
# ~~~~~
# - Proxy specified functions, see "MATCH_STRING" (below)
#   - Use [Interface] to specific which functions to match
#   - Wildcard list of functions to match
#   - Wildcard list of functions to not match
#
# - Support multiple proxy_functions per Proxy object
#   - __init__(..., proxy_function) --> __init__(..., [proxy_function])
#
# - Make @proxy and @remote_proxy decorators
#
# - Aspect Oriented Programming ...
#   - Intercept: logging (flight recorder, tracing for diagnosis)
#   - Intercept: remote function call
#   - Intercept: security access
#   - Intercept: timing (performance)

from inspect import getmembers, isfunction, ismethod
import wrapt

__all__ = ["is_callable", "ProxyAllMethods", "proxy_trace"]


def is_callable(attribute):
    return isfunction(attribute) or ismethod(attribute)

class ProxyAllMethods(wrapt.ObjectProxy):
    def __init__(
        self, proxy_name, actual_object, proxy_function,
        attribute_filter=ismethod, ignore_prefix="_"):

        super(ProxyAllMethods, self).__init__(actual_object)

        def make_closure(actual_function, actual_function_name):
            def closure(*args, **kwargs):
                return proxy_function(
                    proxy_name, actual_object, actual_function,
                    actual_function_name, *args, **kwargs
                )
            return closure

        members = getmembers(actual_object, attribute_filter)
        for name, actual_function in members:
            if ignore_prefix is None or not name.startswith(ignore_prefix):
                closure = make_closure(actual_function, name)
                setattr(self, name, closure)

    def __repr__(self):
        return f"[{self.__module__}.{type(self).__name__} " \
               f"object at {hex(id(self))}]"

def proxy_trace(
    proxy_name, actual_object, actual_function,
    actual_function_name, *args, **kwargs):

    print(f"### Enter: {proxy_name}.{actual_function_name}{args} {kwargs} ###")
    try:
        return actual_function(*args, **kwargs)
    finally:
        print(f"### Exit:  {proxy_name}.{actual_function_name} ###")

class Example:
    def __init__(self, argument_0):
        print(f"           Example.__init__({argument_0})")

    def function_0(self, argument_0, argument_1=None):
        print(f"           Example.function_0({argument_0}, {argument_1})")
        return f"          {argument_0} {argument_1}"

# TODO: @proxy Decorator returns an instance of the "actual_class"
#       Need a factory "actual_class" that creates proxy instances !
#
# def proxy(proxy_function=proxy_trace, *args, **kwargs):
#     def new_function(actual_class):
#         proxy_name = actual_class.__name__
#         instance = actual_class()
#         proxy_object = ProxyAllMethods(proxy_name, instance, proxy_function)
#         return proxy_object
#     return new_function
#
#   from aiko_services.main.proxy import proxy, proxy_trace
#   @proxy
#   class C:
#       def f(self, argument_0):
#           print(f"C.f({argument_0})")



"""Service layer for data source operations."""

from __future__ import annotations

import json
import re
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Literal, Optional, Sequence, Tuple
from uuid import UUID

from highlighter.client import HLClient
from highlighter.client.base_models.base_models import DataSourceType
from highlighter.core.gql_base_model import GQLBaseModel
from highlighter.datasource.models import (
    DataSourceTemplate,
    LocalDataSourceEntry,
    LocalDataSourceFile,
    MatchedDataSource,
    SyncResult,
)
from highlighter.network.mdns import DiscoveredDevice, DiscoveryConfig, discover_devices


class DataSourceService:
    """Service for managing data sources."""

    def __init__(self, client: HLClient):
        self.client = client

    @staticmethod
    def _normalize_mac(mac: Optional[str]) -> Optional[str]:
        if not mac:
            return None
        normalized = re.sub(r"[^0-9a-fA-F]", "", mac).lower()
        return normalized or None

    @staticmethod
    def _normalize_serial(serial: Optional[str]) -> Optional[str]:
        if not serial:
            return None
        normalized = serial.strip().lower()
        return normalized or None

    @staticmethod
    def _raise_on_duplicate_identifier(
        values: Dict[str, List[DataSourceType]],
        identifier_name: str,
    ) -> None:
        duplicates = []
        for key, sources in values.items():
            if len(sources) > 1:
                ids = ", ".join(str(source.id) for source in sources)
                duplicates.append(f"{key} (ids: {ids})")

        if duplicates:
            raise ValueError(
                f"Multiple cloud data sources share the same {identifier_name}: " + "; ".join(duplicates)
            )

    def discover_to_datasources(
        self,
        config: DiscoveryConfig,
        resolve_macs: bool,
        template_obj: Optional[DataSourceTemplate] = None,
        uri_pattern: Optional[str] = None,
        name_pattern: Optional[str] = None,
    ) -> List[LocalDataSourceEntry]:
        """Run mDNS discovery and convert devices to data source entries.

        Args:
            config: Discovery configuration
            resolve_macs: Whether to resolve MAC addresses
            template_obj: Optional template to apply (uses template defaults)
            uri_pattern: Optional URI pattern (only used when template_obj is None)
            name_pattern: Optional name pattern (only used when template_obj is None)

        Returns:
            List of data source entries

        Raises:
            ValueError: If neither template_obj nor uri_pattern is provided
        """
        if template_obj is None and uri_pattern is None:
            raise ValueError("Either template_obj or uri_pattern must be provided")

        devices = discover_devices(config=config, resolve_macs=resolve_macs)

        entries = []
        for device in devices:
            if template_obj:
                # Template mode: use apply_template which handles defaults
                entry = self.apply_template(template_obj, device)
            else:
                # Pattern mode: use custom patterns
                placeholders = {
                    "ip": device.ip or "",
                    "hostname": device.hostname or "",
                    "port": device.port or "",
                    "serial": device.serial or "",
                    "mac": device.mac or "",
                    "service_name": device.service_name or "",
                }

                # Generate source URI and name
                source_uri = uri_pattern.format(**placeholders)
                name = name_pattern.format(**placeholders)

                entry = LocalDataSourceEntry(
                    name=name,
                    source_uri=source_uri,
                    mac=device.mac,
                    serial=device.serial,
                    hostname=device.hostname,
                    ip=device.ip,
                    port=device.port,
                    template=None,
                )

            entries.append(entry)

        return entries

    def load_file(self, file_path: Path) -> LocalDataSourceFile:
        """Load and validate a local data source file."""
        with open(file_path, "r", encoding="utf-8") as f:
            data = json.load(f)
        return LocalDataSourceFile(**data)

    def save_file(self, file_path: Path, data_source_file: LocalDataSourceFile) -> None:
        """Save data source entries to a JSON file."""
        # Update metadata
        data_source_file.metadata.updated_at = datetime.now()

        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(
                data_source_file.model_dump(mode="json", exclude_none=False),
                f,
                indent=2,
                default=str,
            )

    def fetch_from_cloud(self, limit: Optional[int] = None) -> List[DataSourceType]:
        """Query cloud for all data sources with optional limit.

        Tries the paginated `dataSourcesConnection` endpoint; falls back to
        the legacy `dataSources` list if the connection query is unavailable.

        Args:
            limit: Maximum number of data sources to return (None fetches all).
        """

        class PageInfo(GQLBaseModel):
            has_next_page: bool
            end_cursor: Optional[str] = None

        class DataSourceEdge(GQLBaseModel):
            node: DataSourceType

        class DataSourcesConnection(GQLBaseModel):
            edges: List[DataSourceEdge]
            page_info: PageInfo

        page_size = 100
        after = None
        all_sources: List[DataSourceType] = []

        try:
            while True:
                kwargs = {"first": page_size}
                if after:
                    kwargs["after"] = after

                result = self.client.data_sources_connection(return_type=DataSourcesConnection, **kwargs)

                all_sources.extend(edge.node for edge in result.edges)

                if limit and len(all_sources) >= limit:
                    return all_sources[:limit]

                if not result.page_info.has_next_page:
                    break

                after = result.page_info.end_cursor

            return all_sources
        except ValueError as exc:
            # Backward compatibility when the connection query is not supported
            if "dataSourcesConnection" not in str(exc):
                raise

            sources = self.client.data_sources(return_type=List[DataSourceType])
            if limit is not None:
                return sources[:limit]
            return sources

    def get_by_id(self, data_source_id: int) -> Optional[DataSourceType]:
        """Get a single data source by ID."""
        try:
            return self.client.data_source(return_type=DataSourceType, id=data_source_id)
        except Exception:
            return None

    def get_by_uuid(self, data_source_uuid: UUID) -> Optional[DataSourceType]:
        """Get a single data source by UUID."""
        try:
            return self.client.data_source(return_type=DataSourceType, uuid=str(data_source_uuid))
        except Exception:
            return None

    def get_by_name(self, name: str) -> Optional[DataSourceType]:
        """Get a single data source by name.

        Note: Returns the first match if multiple data sources have the same name.
        """
        all_sources = self.fetch_from_cloud()
        for source in all_sources:
            if source.name == name:
                return source
        return None

    def create(
        self,
        name: str,
        source_uri: Optional[str] = None,
        uuid: Optional[UUID] = None,
        serial_number: Optional[str] = None,
        mac_address: Optional[str] = None,
        device_serial_number: Optional[str] = None,
    ) -> DataSourceType:
        """Create a single data source in the cloud.

        `serial_number` is the data source's logical serial identifier.
        `device_serial_number` captures the underlying device/hardware serial when applicable.
        """

        class CreateDataSourcePayload(GQLBaseModel):
            data_source: Optional[DataSourceType]
            errors: List[str]

        kwargs = {"name": name}
        if uuid:
            kwargs["uuid"] = str(uuid)
        if source_uri:
            # GraphQL expects camelCase
            kwargs["sourceUri"] = source_uri
        if serial_number is not None:
            kwargs["serialNumber"] = serial_number
        if mac_address is not None:
            kwargs["macAddress"] = mac_address
        if device_serial_number is not None:
            kwargs["deviceSerialNumber"] = device_serial_number

        response = self.client.create_data_source(return_type=CreateDataSourcePayload, **kwargs)

        if response.errors:
            raise ValueError(f"Failed to create data source: {response.errors}")

        return response.data_source

    def update(
        self,
        data_source_id: Optional[int] = None,
        data_source_uuid: Optional[UUID] = None,
        name: Optional[str] = None,
        source_uri: Optional[str] = None,
        serial_number: Optional[str] = None,
        mac_address: Optional[str] = None,
        device_serial_number: Optional[str] = None,
    ) -> DataSourceType:
        """Update an existing data source in the cloud.

        `serial_number` is the data source's logical serial identifier.
        `device_serial_number` captures the underlying device/hardware serial when applicable.
        """

        class UpdateDataSourcePayload(GQLBaseModel):
            data_source: Optional[DataSourceType]
            errors: List[str]

        kwargs = {}
        if data_source_id:
            kwargs["id"] = data_source_id
        elif data_source_uuid:
            kwargs["uuid"] = str(data_source_uuid)

        if name:
            kwargs["name"] = name
        if source_uri:
            kwargs["sourceUri"] = source_uri
        if serial_number is not None:
            kwargs["serialNumber"] = serial_number
        if mac_address is not None:
            kwargs["macAddress"] = mac_address
        if device_serial_number is not None:
            kwargs["deviceSerialNumber"] = device_serial_number

        response = self.client.update_data_source(return_type=UpdateDataSourcePayload, **kwargs)

        if response.errors:
            raise ValueError(f"Failed to update data source: {response.errors}")

        return response.data_source

    def destroy(self, data_source_id: Optional[int] = None, data_source_uuid: Optional[UUID] = None) -> bool:
        """Delete a data source from the cloud."""

        class DeleteDataSourcePayload(GQLBaseModel):
            success: bool
            errors: List[str]

        kwargs = {}
        if data_source_id:
            kwargs["id"] = data_source_id
        elif data_source_uuid:
            kwargs["uuid"] = str(data_source_uuid)

        response = self.client.delete_data_source(return_type=DeleteDataSourcePayload, **kwargs)

        if response.errors:
            raise ValueError(f"Failed to delete data source: {response.errors}")

        return response.success

    def match_entries(
        self,
        local_entries: List[LocalDataSourceEntry],
        cloud_sources: List[DataSourceType],
        match_by: Literal["id", "uuid", "name", "source_uri", "mac", "serial", "auto"] = "auto",
    ) -> List[MatchedDataSource]:
        """Match local data source entries against cloud data sources."""
        matches = []

        # Create lookup dictionaries for cloud sources
        by_id: Dict[int, DataSourceType] = {s.id: s for s in cloud_sources}
        by_uuid: Dict[UUID, DataSourceType] = {s.uuid: s for s in cloud_sources}
        by_name: Dict[str, DataSourceType] = {s.name: s for s in cloud_sources}
        by_source_uri: Dict[str, DataSourceType] = {
            getattr(s, "source_uri", None): s for s in cloud_sources if getattr(s, "source_uri", None)
        }
        by_mac: Dict[str, List[DataSourceType]] = {}
        by_serial: Dict[str, List[DataSourceType]] = {}

        for source in cloud_sources:
            mac = self._normalize_mac(getattr(source, "mac_address", None))
            if mac:
                by_mac.setdefault(mac, []).append(source)

            serial = self._normalize_serial(getattr(source, "serial_number", None))
            if serial:
                by_serial.setdefault(serial, []).append(source)

        self._raise_on_duplicate_identifier(by_mac, "mac_address")
        self._raise_on_duplicate_identifier(by_serial, "serial_number")

        by_mac_single: Dict[str, DataSourceType] = {key: sources[0] for key, sources in by_mac.items()}
        by_serial_single: Dict[str, DataSourceType] = {key: sources[0] for key, sources in by_serial.items()}

        for entry in local_entries:
            matched = MatchedDataSource(local=entry)
            local_mac = self._normalize_mac(entry.mac)
            local_serial = self._normalize_serial(entry.serial)

            if match_by == "auto":
                # Try matching in order: id -> uuid -> mac -> serial -> source_uri -> name
                if entry.id and entry.id in by_id:
                    matched.cloud = by_id[entry.id]
                    matched.match_type = "id"
                    matched.is_matched = True
                elif entry.uuid and entry.uuid in by_uuid:
                    matched.cloud = by_uuid[entry.uuid]
                    matched.match_type = "uuid"
                    matched.is_matched = True
                elif local_mac and local_mac in by_mac_single:
                    matched.cloud = by_mac_single[local_mac]
                    matched.match_type = "mac"
                    matched.is_matched = True
                elif local_serial and local_serial in by_serial_single:
                    matched.cloud = by_serial_single[local_serial]
                    matched.match_type = "serial"
                    matched.is_matched = True
                elif entry.source_uri and entry.source_uri in by_source_uri:
                    matched.cloud = by_source_uri[entry.source_uri]
                    matched.match_type = "source_uri"
                    matched.is_matched = True
                elif entry.name in by_name:
                    matched.cloud = by_name[entry.name]
                    matched.match_type = "name"
                    matched.is_matched = True
            elif match_by == "id" and entry.id and entry.id in by_id:
                matched.cloud = by_id[entry.id]
                matched.match_type = "id"
                matched.is_matched = True
            elif match_by == "uuid" and entry.uuid and entry.uuid in by_uuid:
                matched.cloud = by_uuid[entry.uuid]
                matched.match_type = "uuid"
                matched.is_matched = True
            elif match_by == "mac" and local_mac and local_mac in by_mac_single:
                matched.cloud = by_mac_single[local_mac]
                matched.match_type = "mac"
                matched.is_matched = True
            elif match_by == "serial" and local_serial and local_serial in by_serial_single:
                matched.cloud = by_serial_single[local_serial]
                matched.match_type = "serial"
                matched.is_matched = True
            elif match_by == "source_uri" and entry.source_uri and entry.source_uri in by_source_uri:
                matched.cloud = by_source_uri[entry.source_uri]
                matched.match_type = "source_uri"
                matched.is_matched = True
            elif match_by == "name" and entry.name in by_name:
                matched.cloud = by_name[entry.name]
                matched.match_type = "name"
                matched.is_matched = True

            matches.append(matched)

        return matches

    def import_to_cloud(
        self,
        entries: List[LocalDataSourceEntry],
        create_missing: bool = False,
        update_existing: bool = False,
        match_by: Literal["id", "uuid", "name", "source_uri", "mac", "serial", "auto"] = "auto",
    ) -> Tuple[List[LocalDataSourceEntry], SyncResult]:
        """Import data sources to cloud, optionally creating or updating."""
        cloud_sources = self.fetch_from_cloud()
        matches = self.match_entries(entries, cloud_sources, match_by)

        result = SyncResult()
        updated_entries = []

        for match in matches:
            try:
                if match.is_matched:
                    result.matched += 1
                    # Update cloud ID/UUID in local entry
                    match.local.id = match.cloud.id
                    match.local.uuid = match.cloud.uuid

                    if match.match_type == "source_uri" and create_missing and not update_existing:
                        result.skipped_duplicates += 1
                        result.skipped_reasons.append(
                            f"Skipping create: source_uri '{match.local.source_uri}' already exists in cloud "
                            f"(id={match.cloud.id}, name={match.cloud.name})"
                        )

                    if update_existing:
                        # Update the cloud entry
                        self.update(
                            data_source_id=match.cloud.id,
                            name=match.local.name,
                            source_uri=match.local.source_uri,
                            serial_number=match.local.serial,
                            mac_address=match.local.mac,
                        )
                        result.updated += 1

                    updated_entries.append(match.local)
                else:
                    # No match found
                    if create_missing:
                        # Create new data source (no source_uri dedupe unless matched above)
                        created = self.create(
                            name=match.local.name,
                            source_uri=match.local.source_uri,
                            serial_number=match.local.serial,
                            mac_address=match.local.mac,
                        )
                        match.local.id = created.id
                        match.local.uuid = created.uuid
                        result.created += 1

                    updated_entries.append(match.local)

            except Exception as e:
                result.failed += 1
                result.errors.append(f"Failed to process {match.local.name}: {str(e)}")
                updated_entries.append(match.local)

        return updated_entries, result

    def apply_template(
        self, template: DataSourceTemplate, device: DiscoveredDevice, requested_name: Optional[str] = None
    ) -> LocalDataSourceEntry:
        """Apply a template to generate a data source entry from a discovered device.

        Applies template defaults and persists the resolved values in the entry fields.
        Validates required fields according to template definition.

        Args:
            template: Template with field requirements and defaults
            device: Discovered device

        Raises:
            ValueError: If required template fields are missing from device
        """
        # Map device attributes to template field names
        device_values = {
            "ip": device.ip,
            "hostname": device.hostname,
            "port": device.port,
            "serial": device.serial,
            "mac": device.mac,
            "name": device.service_name,
        }

        # Validate required fields (skip generated fields like source_uri)
        generated_fields = {"source_uri"}  # These are generated by the template, not from device
        missing_fields = []
        for field_name, field_def in template.fields.items():
            if field_def.required and field_name not in generated_fields:
                value = device_values.get(field_name)
                # Check if value is None or empty string
                if not value:
                    # Check if there's a default in the template
                    has_default = field_def.default is not None
                    # Check if there's a default_port/default_path for specific fields
                    has_template_default = (field_name == "port" and template.default_port) or (
                        field_name == "path" and template.default_path
                    )
                    if not has_default and not has_template_default:
                        missing_fields.append(field_name)

        if missing_fields:
            device_desc = device.service_name or device.ip or "unknown"
            raise ValueError(
                f"Device '{device_desc}' is missing required template fields: {', '.join(missing_fields)}"
            )

        # Resolve values with template defaults
        resolved_port = device.port or (str(template.default_port) if template.default_port else None)

        # Build placeholders for URI/name generation
        base_name = requested_name or device.service_name or ""

        placeholders = {
            "ip": device.ip or "",
            "hostname": device.hostname or "",
            "port": resolved_port or "",
            "serial": device.serial or "",
            "mac": device.mac or "",
            "path": template.default_path or "",
            "name": base_name,
            "service_name": base_name,
        }

        source_uri = template.source_uri_pattern.format(**placeholders)
        if template.name_pattern:
            entry_name = template.name_pattern.format(**placeholders)
        else:
            entry_name = base_name or "Unknown Device"

        # Persist the resolved port value (not the original device.port which may be None)
        return LocalDataSourceEntry(
            name=entry_name,
            source_uri=source_uri,
            mac=device.mac,
            serial=device.serial,
            hostname=device.hostname,
            ip=device.ip,
            port=resolved_port,  # Use resolved value with template defaults
            template=template.name,
        )

    def load_template(self, template_name: str) -> Optional[DataSourceTemplate]:
        """Load a template by name from built-in or a user-provided JSON file path."""
        # First check built-in templates
        builtin_path = Path(__file__).parent / "templates" / f"{template_name}.json"
        if builtin_path.exists():
            with open(builtin_path, "r", encoding="utf-8") as f:
                data = json.load(f)
            return DataSourceTemplate(**data)

        # Then check user-provided path
        candidate_paths = []
        raw_path = Path(template_name)
        candidate_paths.append(raw_path)
        if raw_path.suffix == "":
            candidate_paths.append(raw_path.with_suffix(".json"))

        for path in candidate_paths:
            if path.exists():
                with open(path, "r", encoding="utf-8") as f:
                    data = json.load(f)
                return DataSourceTemplate(**data)

        return None

    def list_templates(self) -> List[str]:
        """List available template names."""
        templates_dir = Path(__file__).parent / "templates"
        if not templates_dir.exists():
            return []

        return [p.stem for p in templates_dir.glob("*.json")]



import logging
import time
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID

import numpy as np
from pydantic import BaseModel, Field, model_validator

from highlighter.client import DatumSource
from highlighter.client.gql_client import HLClient
from highlighter.client.training_config import TrainingConfigType, get_training_config
from highlighter.client.training_runs import get_training_run_artefact
from highlighter.core import LabeledUUID
from highlighter.core.const import OBJECT_CLASS_ATTRIBUTE_UUID
from highlighter.core.exceptions import OptionalPackageMissingError
from highlighter.predictors.output_category_descriptor import (
    EnumDescriptorWithConfidenceThreshold,
    OutputCategoryDescriptor,
)

try:
    import cv2
except ModuleNotFoundError as _:
    raise OptionalPackageMissingError("cv2", "cv2")

try:
    import torch
    import torch.nn.functional as F
except ModuleNotFoundError as _:
    raise OptionalPackageMissingError("torch", "torch")


from highlighter.client import TrainingRunArtefactType
from highlighter.client.base_models.annotation import Annotation
from highlighter.client.base_models.observation import Observation
from highlighter.core.data_models import DataSample

from .onnx_predictor import OnnxPredictor

__all__ = ["OnnxYoloV8"]

LOG = logging.getLogger(__name__)


class OnnxYoloV8:

    def __init__(
        self,
        onnx_predictor: Union[OnnxPredictor, str, Path],
        num_classes: int,
        class_lookup: Optional[Dict[int | str, OutputCategoryDescriptor]] = None,
        conf_thresh: float = 0.1,
        nms_iou_thresh: float = 0.5,
        nms_class_agnostic: bool = False,
        is_absolute: bool = True,
        **kwargs,
    ):
        """
        Args:
            num_classes: The number of output classes of the predictor
            class_lookup: Optional. If set then only classes in the dict will be returned. If not set, the
              the Annotation's ObjectClass Observation value with be UUID(int=class_id) with no 'human readable' label.
              Set with one of:
                - [object_class_uuid, object_class_label]
                - [object_class_uuid, object_class_label, confidence_threshold]
                - [attribute_id, attribute_label, enum_id, enum_label]
                - [attribute_id, attribute_label, enum_id, enum_label, confidence_threshold]
            conf_thresh: Only return detections with confidence >= conf_thresh
            nms_iou_thresh: Detections with an IOU >= nms_iou_thresh and the same class_id will be considered the same entity and the
            one with the max confidence will be used
            is_absolute: True if the model returns detections in absolute pixel values, else set to false.
            kwargs: Optional args to pass to OnnxPredictor
        """

        if isinstance(onnx_predictor, OnnxPredictor):
            self._onnx_predictor = onnx_predictor
        else:
            self._onnx_predictor = OnnxPredictor(onnx_predictor, **kwargs)

        # Numpy dtype: support both FP32 and FP16 onnx model
        self.ndtype = (
            np.half if self._onnx_predictor.sess.get_inputs()[0].type == "tensor(float16)" else np.single
        )
        self.num_classes = num_classes

        if class_lookup is not None:
            # Ensure keys are int so that class lookup works.
            # JSON doesn't allow numbers as keys so we may receive strings.
            # And, while we're at it we also any per-class confidence thresholds
            # or default to conf_thresh
            self.class_lookup = {}
            for k, v in class_lookup.items():
                self.class_lookup[int(k)] = (
                    EnumDescriptorWithConfidenceThreshold.from_output_category_descriptor(
                        v, default_confidence_threshold=conf_thresh
                    )
                )
        else:
            self.class_lookup = {
                i: EnumDescriptorWithConfidenceThreshold(
                    attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                    attribute_label="object_class",
                    enum_id=UUID(int=i),
                    enum_label="",
                    confidence_threshold=conf_thresh,
                )
                for i in range(self.num_classes)
            }

        self.nms_box_conf = min([v.confidence_threshold for v in self.class_lookup.values()])
        # Boxes with iou < this will be filterd out during NMS
        self.nms_iou_thr = nms_iou_thresh
        self.nms_class_agnostic = nms_class_agnostic
        self.is_absolute = is_absolute

        # Get model width and height(YOLOv8-seg only has one input)
        self.model_height, self.model_width = [x.shape for x in self._onnx_predictor.sess.get_inputs()][0][
            -2:
        ]
        self.use_high_res_masks = False  # see:

        onnx_model_input_shape = self._onnx_predictor.sess.get_inputs()[0].shape

        if (len(onnx_model_input_shape) != 4) or (onnx_model_input_shape[0] != 1):
            raise NotImplementedError(
                f"{self.__class__.__name__} can only deal with onnx models "
                "with static input shapes [batch=1, chan=3, h, w] "
                f"got: {onnx_model_input_shape}."
            )

    @classmethod
    def load_from_training_run(
        cls, training_run_id: int, artefact_id: Union[UUID, str], cache_dir: Optional[Path] = None, **kwargs
    ) -> "OnnxYoloV8":
        if cache_dir:
            onnx_file = cache_dir / f"{training_run_id}-{artefact_id}.onnx14"
            training_config_path = cache_dir / f"{training_run_id}-{artefact_id}.yaml"

            if not onnx_file.exists():
                get_training_run_artefact(
                    HLClient.get_client(),
                    str(artefact_id),
                    download_file_url=True,
                    file_url_save_path=str(onnx_file),
                )

            if not training_config_path.exists():
                training_config = get_training_config(HLClient.get_client(), training_run_id)
                training_config.dump("yaml", training_config_path)
            else:
                training_config = TrainingConfigType.from_yaml(training_config_path)
        else:
            onnx_file = get_training_run_artefact(
                HLClient.get_client(),
                str(artefact_id),
                download_file_url=True,
            )
            training_config = get_training_config(HLClient.get_client(), training_run_id)

        class_lookup = {
            p.position: EnumDescriptorWithConfidenceThreshold(
                attribute_id=p.taxon[0].entity_attribute_id,
                attribute_label=p.taxon[0].entity_attribute_name,
                enum_id=p.taxon[0].entity_attribute_enum_id,
                enum_label=p.taxon[0].entity_attribute_enum_value,
                confidence_threshold=p.conf_thresh,
            )
            for p in training_config.input_output_schema.get_head_model_outputs(0)
        }

        return cls(OnnxPredictor(str(onnx_file)), len(class_lookup), class_lookup=class_lookup, **kwargs)

    @classmethod
    def load_from_artefact(cls, training_run_artefact: TrainingRunArtefactType) -> "OnnxYoloV8":
        onnx_predictor = OnnxPredictor(training_run_artefact.file_url)
        num_classes = len(training_run_artefact.inference_config["model_schema"]["model_outputs"])
        return cls(onnx_predictor, num_classes=num_classes)

    def predict(self, data_samples: List[DataSample], **_) -> List[List[Annotation]]:
        batch_output: List[Annotation] = []
        for data_sample in data_samples:
            img = data_sample.content
            img_pre = preprocess(img, self.model_height, self.model_width, self.ndtype)
            raw_ort_output = self._onnx_predictor.predict_batch(img_pre)
            if len(raw_ort_output[0].shape) == 2:
                class_indices = np.argmax(raw_ort_output[0], axis=1)
                confs = np.max(raw_ort_output[0], axis=1)
                locations = [(0, 0) + data_sample.wh for _ in confs]
                image_annotations = self._make_annotations(
                    Annotation.from_left_top_right_bottom_box, class_indices, confs, locations, data_sample
                )

            else:
                # Need to fork here and run _postprocess depending if we're
                # dealing with a classification responce or det/seg
                image_annotations = self._postprocess(
                    raw_ort_output,
                    img_pre.shape[-2:],
                    data_sample,
                    self.is_absolute,
                )

            batch_output.append(image_annotations)
        return batch_output

    def _make_annotations(
        self, make_anno_fn, class_indices, confs, locations, data_sample
    ) -> List[Annotation]:
        LOG.verbose(f"{self.__class__.__name__} confs: {confs}")
        LOG.verbose(f"{self.__class__.__name__} class_indices: {class_indices}")
        LOG.verbose(f"{self.__class__.__name__} locations: {locations}")
        annotations: List[Annotation] = []
        created_records = []
        dropped_counts: Dict[str, int] = {"unknown_class": 0, "below_threshold": 0, "zero_area": 0}

        def _format_location(loc):
            if isinstance(loc, np.ndarray):
                if loc.ndim > 1:
                    return f"array shape={loc.shape}"
                loc = loc.tolist()
            if isinstance(loc, (list, tuple)) and len(loc) == 4:
                x0, y0, x1, y1 = loc
                return f"({x0:.2f},{y0:.2f})-({x1:.2f},{y1:.2f})"
            return str(loc)

        category_descriptors = [self.class_lookup.get(class_index, None) for class_index in class_indices]

        for category_descriptor, conf, loc in zip(category_descriptors, confs, locations):
            if category_descriptor is None:
                dropped_counts["unknown_class"] += 1
                LOG.debug(
                    f"Dropping detection with unknown category={category_descriptor}, loc={_format_location(loc)}"
                )
                continue

            if conf < category_descriptor.confidence_threshold:
                dropped_counts["below_threshold"] += 1
                LOG.debug(
                    f"Dropping detection below threshold (conf={conf:.3f} < {category_descriptor.confidence_threshold:.3f}) "
                    f"category={category_descriptor}, loc={_format_location(loc)}"
                )
                continue

            # Drop degenerate boxes that would become invalid polygons
            if isinstance(loc, (list, tuple, np.ndarray)) and len(loc) == 4:
                x0, y0, x1, y1 = loc
                if (x1 - x0) <= 0 or (y1 - y0) <= 0:
                    dropped_counts["zero_area"] += 1
                    LOG.debug(
                        f"Dropping zero-area box loc={_format_location(loc)} "
                        f"category={category_descriptor}, conf={conf:.3f}"
                    )
                    continue

            obs = Observation(
                attribute_id=LabeledUUID(
                    category_descriptor.attribute_id,
                    label=category_descriptor.attribute_label,
                ),
                value=LabeledUUID(category_descriptor.enum_id, label=category_descriptor.enum_label),
                datum_source=DatumSource(confidence=conf),
            )

            anno = make_anno_fn(
                loc,
                obs.datum_source.confidence,
                observations=[obs],
                data_sample=data_sample,
            )
            annotations.append(anno)
            created_records.append(
                {
                    "category": category_descriptor,
                    "conf": conf,
                    "loc": loc,
                }
            )

        LOG.verbose(
            f"{self.__class__.__name__} created {len(annotations)} annotations; dropped={dropped_counts} "
            f"locations={[ _format_location(rec['loc']) for rec in created_records ]}"
        )
        for rec in created_records:
            LOG.debug(
                f"Annotation: category='{rec['category']}'), "
                f"conf={rec['conf']:.3f}, loc={_format_location(rec['loc'])}"
            )
        return annotations

    def _postprocess(
        self,
        batch_ort_output,
        preproc_img_hw: Tuple[int, int],
        data_sample,
        is_absolute: bool,
    ) -> List[Annotation]:
        orig_img_hw = data_sample.wh[-1::-1]

        annotations: List[Annotation] = []
        if len(batch_ort_output) == 1:  # detection output
            if batch_ort_output[0].shape[0] != 1:
                # ToDo: Deal with batches, first need to figure out how to export
                # yolov8 with variable batch size
                raise NotImplementedError("Can only handel a batch size of one for now")

            # ultralytics/models/yolo/detect/predict.py (commit: b617e131bdf4ba6652a55f71b95e2c4c00e2ee59)
            preds = non_max_suppression(
                torch.Tensor(batch_ort_output[0]),
                self.nms_box_conf,
                self.nms_iou_thr,
                nc=self.num_classes,
                agnostic=self.nms_class_agnostic,
            )[0]

            if not is_absolute:
                LOG.verbose(f"{self.__class__.__name__} scaling relative boxes (before): {preds[:, :4]}")
                preds[:, [0, 2]] = preds[:, [0, 2]] * preproc_img_hw[1]
                preds[:, [1, 3]] = preds[:, [1, 3]] * preproc_img_hw[0]

            LOG.verbose(f"{self.__class__.__name__} preproc_img_hw: {preproc_img_hw}")
            LOG.verbose(f"{self.__class__.__name__} orig_img_hw: {orig_img_hw}")
            LOG.verbose(f"{self.__class__.__name__} preds (pre scale_boxes): {preds[:, :4]}")
            preds[:, :4] = scale_boxes(preproc_img_hw, preds[:, :4], orig_img_hw)
            preds = preds.numpy()
            trlb_boxes = preds[:, :4]
            confs = preds[:, 4]
            class_indices = preds[:, 5].astype(int)

            annotations = self._make_annotations(
                Annotation.from_left_top_right_bottom_box, class_indices, confs, trlb_boxes, data_sample
            )

            # FIXME: DataSample, how do we associate Annotations to a data files

        elif len(batch_ort_output) == 2:  # segmentation output
            # Rename to make it easier to follow the reference implementation
            # ultralytics/models/yolo/segment/predict.py (commit: b617e131bdf4ba6652a55f71b95e2c4c00e2ee59)
            preds = batch_ort_output
            p = non_max_suppression(
                torch.Tensor(preds[0]),
                self.nms_box_conf,
                self.nms_iou_thr,
                nc=self.num_classes,
                agnostic=self.nms_class_agnostic,
            )

            proto = (
                preds[1][-1] if isinstance(preds[1], tuple) else preds[1]
            )  # tuple if PyTorch model or array if exported

            if len(p) != 1:
                # ToDo: Deal with batches, first need to figure out how to export
                # yolov8 with variable batch size
                raise NotImplementedError("Can only handel a batch size of one for now")

            pred = p[0]
            proto = proto[0]

            if len(pred):
                masks = process_mask(
                    torch.Tensor(proto), pred[:, 6:], pred[:, :4], preproc_img_hw, upsample=True
                )  # HWC
                masks = scale_masks(masks[None], orig_img_hw)[0].numpy()
                pred[:, :4] = scale_boxes(preproc_img_hw, pred[:, :4], orig_img_hw)

                confs = pred[:, 4].numpy()
                class_indices = pred[:, 5].numpy()

                annotations = self._make_annotations(
                    Annotation.from_mask, class_indices, confs, masks, data_sample
                )

        else:
            raise ValueError("Invalid batch_ort_output")

        return annotations


# Adapted from: ultralytics/data/augment.py (commit: b617e131bdf4ba6652a55f71b95e2c4c00e2ee59)
def preprocess(img, new_height, new_width, ndtype):
    """
    Pre-processes the input image.

    Args:
        img (Numpy.ndarray): image about to be processed.

    Returns:
        img_process (Numpy.ndarray): image preprocessed for inference.
    """

    # Resize and pad input image using letterbox() (Borrowed from Ultralytics)
    shape = img.shape[:2]  # original image shape
    new_shape = (new_height, new_width)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    ratio = r, r
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    pad_w, pad_h = (new_shape[1] - new_unpad[0]) / 2, (new_shape[0] - new_unpad[1]) / 2  # wh padding
    if shape[::-1] != new_unpad:  # resize
        img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(pad_h - 0.1)), int(round(pad_h + 0.1))
    left, right = int(round(pad_w - 0.1)), int(round(pad_w + 0.1))
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=(114, 114, 114))

    # Transforms: HWC to CHW -> BGR to RGB -> div(255) -> contiguous -> add axis(optional)
    # img = np.ascontiguousarray(np.einsum("HWC->CHW", img)[::-1], dtype=ndtype) / 255.0
    img = np.ascontiguousarray(np.einsum("HWC->CHW", img), dtype=ndtype) / 255.0

    img_process = img[None] if len(img.shape) == 3 else img
    return img_process


# Taken from: ultralytics/utils/metrics.py (commit: b617e131bdf4ba6652a55f71b95e2c4c00e2ee59)
def _get_covariance_matrix(boxes):
    """
    Generating covariance matrix from obbs.

    Args:
        boxes (torch.Tensor): A tensor of shape (N, 5) representing rotated bounding boxes, with xywhr format.

    Returns:
        (torch.Tensor): Covariance metrixs corresponding to original rotated bounding boxes.
    """
    # Gaussian bounding boxes, ignore the center points (the first two columns) because they are not needed here.
    gbbs = torch.cat((boxes[:, 2:4].pow(2) / 12, boxes[:, 4:]), dim=-1)
    a, b, c = gbbs.split(1, dim=-1)
    cos = c.cos()
    sin = c.sin()
    cos2 = cos.pow(2)
    sin2 = sin.pow(2)
    return a * cos2 + b * sin2, a * sin2 + b * cos2, (a - b) * cos * sin


# Taken from: ultralytics/utils/metrics.py (commit: b617e131bdf4ba6652a55f71b95e2c4c00e2ee59)
def batch_probiou(obb1, obb2, eps=1e-7):
    """
    Calculate the prob IoU between oriented bounding boxes, https://arxiv.org/pdf/2106.06072v1.pdf.

    Args:
        obb1 (torch.Tensor | np.ndarray): A tensor of shape (N, 5) representing ground truth obbs, with xywhr format.
        obb2 (torch.Tensor | np.ndarray): A tensor of shape (M, 5) representing predicted obbs, with xywhr format.
        eps (float, optional): A small value to avoid division by zero. Defaults to 1e-7.

    Returns:
        (torch.Tensor): A tensor of shape (N, M) representing obb similarities.
    """
    obb1 = torch.from_numpy(obb1) if isinstance(obb1, np.ndarray) else obb1
    obb2 = torch.from_numpy(obb2) if isinstance(obb2, np.ndarray) else obb2

    x1, y1 = obb1[..., :2].split(1, dim=-1)
    x2, y2 = (x.squeeze(-1)[None] for x in obb2[..., :2].split(1, dim=-1))
    a1, b1, c1 = _get_covariance_matrix(obb1)
    a2, b2, c2 = (x.squeeze(-1)[None] for x in _get_covariance_matrix(obb2))

    t1 = (
        ((a1 + a2) * (y1 - y2).pow(2) + (b1 + b2) * (x1 - x2).pow(2))
        / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)
    ) * 0.25
    t2 = (((c1 + c2) * (x2 - x1) * (y1 - y2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)) * 0.5
    t3 = (
        ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2))
        / (4 * ((a1 * b1 - c1.pow(2)).clamp_(0) * (a2 * b2 - c2.pow(2)).clamp_(0)).sqrt() + eps)
        + eps
    ).log() * 0.5
    bd = (t1 + t2 + t3).clamp(eps, 100.0)
    hd = (1.0 - (-bd).exp() + eps).sqrt()
    return 1 - hd


# Taken from: ultralytics/utils/ops.py (commit: b617e131bdf4ba6652a55f71b95e2c4c00e2ee59)
def nms_rotated(boxes, scores, threshold=0.45):
    """
    NMS for obbs, powered by probiou and fast-nms.

    Args:
        boxes (torch.Tensor): (N, 5), xywhr.
        scores (torch.Tensor): (N, ).
        threshold (float): IoU threshold.

    Returns:
    """
    if len(boxes) == 0:
        return np.empty((0,), dtype=np.int8)
    sorted_idx = torch.argsort(scores, descending=True)
    boxes = boxes[sorted_idx]
    ious = batch_probiou(boxes, boxes).triu_(diagonal=1)
    pick = torch.nonzero(ious.max(dim=0)[0] < threshold).squeeze_(-1)
    return sorted_idx[pick]


# Taken from: ultralytics/utils/ops.py (commit: b617e131bdf4ba6652a55f71b95e2c4c00e2ee59)
def xywh2xyxy(x):
    """
    Convert bounding box coordinates from (x, y, width, height) format to (x1, y1, x2, y2) format where (x1, y1) is the
    top-left corner and (x2, y2) is the bottom-right corner.

    Args:
        x (np.ndarray | torch.Tensor): The input bounding box coordinates in (x, y, width, height) format.

    Returns:
        y (np.ndarray | torch.Tensor): The bounding box coordinates in (x1, y1, x2, y2) format.
    """
    assert x.shape[-1] == 4, f"input shape last dimension expected 4 but input shape is {x.shape}"
    y = torch.empty_like(x) if isinstance(x, torch.Tensor) else np.empty_like(x)  # faster than clone/copy
    dw = x[..., 2] / 2  # half-width
    dh = x[..., 3] / 2  # half-height
    y[..., 0] = x[..., 0] - dw  # top left x
    y[..., 1] = x[..., 1] - dh  # top left y
    y[..., 2] = x[..., 0] + dw  # bottom right x
    y[..., 3] = x[..., 1] + dh  # bottom right y
    return y


# Taken from: ultralytics/utils/ops.py (commit: b617e131bdf4ba6652a55f71b95e2c4c00e2ee59)
def non_max_suppression(
    prediction,
    conf_thres=0.25,
    iou_thres=0.45,
    classes=None,
    agnostic=False,
    multi_label=False,
    labels=(),
    max_det=300,
    nc=0,  # number of classes (optional)
    max_time_img=0.05,
    max_nms=30000,
    max_wh=7680,
    in_place=True,
    rotated=False,
):
    """
    Perform non-maximum suppression (NMS) on a set of boxes, with support for masks and multiple labels per box.

    Args:
        prediction (torch.Tensor): A tensor of shape (batch_size, num_classes + 4 + num_masks, num_boxes)
            containing the predicted boxes, classes, and masks. The tensor should be in the format
            output by a model, such as YOLO.
        conf_thres (float): The confidence threshold below which boxes will be filtered out.
            Valid values are between 0.0 and 1.0.
        iou_thres (float): The IoU threshold below which boxes will be filtered out during NMS.
            Valid values are between 0.0 and 1.0.
        classes (List[int]): A list of class indices to consider. If None, all classes will be considered.
        agnostic (bool): If True, the model is agnostic to the number of classes, and all
            classes will be considered as one.
        multi_label (bool): If True, each box may have multiple labels.
        labels (List[List[Union[int, float, torch.Tensor]]]): A list of lists, where each inner
            list contains the apriori labels for a given image. The list should be in the format
            output by a dataloader, with each label being a tuple of (class_index, x1, y1, x2, y2).
        max_det (int): The maximum number of boxes to keep after NMS.
        nc (int, optional): The number of classes output by the model. Any indices after this will be considered masks.
        max_time_img (float): The maximum time (seconds) for processing one image.
        max_nms (int): The maximum number of boxes into torchvision.ops.nms().
        max_wh (int): The maximum box width and height in pixels.
        in_place (bool): If True, the input prediction tensor will be modified in place.

    Returns:
        (List[torch.Tensor]): A list of length batch_size, where each element is a tensor of
            shape (num_boxes, 6 + num_masks) containing the kept boxes, with columns
            (x1, y1, x2, y2, confidence, class, mask1, mask2, ...).
    """
    import torchvision  # scope for faster 'import ultralytics'

    # Checks
    assert (
        0 <= conf_thres <= 1
    ), f"Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0"
    assert 0 <= iou_thres <= 1, f"Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0"
    if isinstance(
        prediction, (list, tuple)
    ):  # YOLOv8 model in validation model, output = (inference_out, loss_out)
        prediction = prediction[0]  # select only inference output

    bs = prediction.shape[0]  # batch size
    nc = nc or (prediction.shape[1] - 4)  # number of classes
    nm = prediction.shape[1] - nc - 4
    mi = 4 + nc  # mask start index
    xc = prediction[:, 4:mi].amax(1) > conf_thres  # candidates

    # Settings
    # min_wh = 2  # (pixels) minimum box width and height
    time_limit = 2.0 + max_time_img * bs  # seconds to quit after
    multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)

    prediction = prediction.transpose(-1, -2)  # shape(1,84,6300) to shape(1,6300,84)
    if not rotated:
        if in_place:
            prediction[..., :4] = xywh2xyxy(prediction[..., :4])  # xywh to xyxy
        else:
            prediction = torch.cat(
                (xywh2xyxy(prediction[..., :4]), prediction[..., 4:]), dim=-1
            )  # xywh to xyxy

    t = time.time()
    output = [torch.zeros((0, 6 + nm), device=prediction.device)] * bs
    for xi, x in enumerate(prediction):  # image index, image inference
        # Apply constraints
        # x[((x[:, 2:4] < min_wh) | (x[:, 2:4] > max_wh)).any(1), 4] = 0  # width-height
        x = x[xc[xi]]  # confidence

        # Cat apriori labels if autolabelling
        if labels and len(labels[xi]) and not rotated:
            lb = labels[xi]
            v = torch.zeros((len(lb), nc + nm + 4), device=x.device)
            v[:, :4] = xywh2xyxy(lb[:, 1:5])  # box
            v[range(len(lb)), lb[:, 0].long() + 4] = 1.0  # cls
            x = torch.cat((x, v), 0)

        # If none remain process next image
        if not x.shape[0]:
            continue

        # Detections matrix nx6 (xyxy, conf, cls)
        box, cls, mask = x.split((4, nc, nm), 1)

        if multi_label:
            i, j = torch.where(cls > conf_thres)
            x = torch.cat((box[i], x[i, 4 + j, None], j[:, None].float(), mask[i]), 1)
        else:  # best class only
            conf, j = cls.max(1, keepdim=True)
            x = torch.cat((box, conf, j.float(), mask), 1)[conf.view(-1) > conf_thres]

        # Filter by class
        if classes is not None:
            x = x[(x[:, 5:6] == torch.tensor(classes, device=x.device)).any(1)]

        # Check shape
        n = x.shape[0]  # number of boxes
        if not n:  # no boxes
            continue
        if n > max_nms:  # excess boxes
            x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence and remove excess boxes

        # Batched NMS
        c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
        scores = x[:, 4]  # scores
        if rotated:
            boxes = torch.cat((x[:, :2] + c, x[:, 2:4], x[:, -1:]), dim=-1)  # xywhr
            i = nms_rotated(boxes, scores, iou_thres)
        else:
            boxes = x[:, :4] + c  # boxes (offset by class)
            i = torchvision.ops.nms(boxes, scores, iou_thres)  # NMS
        i = i[:max_det]  # limit detections

        # # Experimental
        # merge = False  # use merge-NMS
        # if merge and (1 < n < 3E3):  # Merge NMS (boxes merged using weighted mean)
        #     # Update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
        #     from .metrics import box_iou
        #     iou = box_iou(boxes[i], boxes) > iou_thres  # IoU matrix
        #     weights = iou * scores[None]  # box weights
        #     x[i, :4] = torch.mm(weights, x[:, :4]).float() / weights.sum(1, keepdim=True)  # merged boxes
        #     redundant = True  # require redundant detections
        #     if redundant:
        #         i = i[iou.sum(1) > 1]  # require redundancy

        output[xi] = x[i]
        if (time.time() - t) > time_limit:
            LOG.warning(f"WARNING ⚠️ NMS time limit {time_limit:.3f}s exceeded")
            break  # time limit exceeded

    return output


def scale_boxes(img1_shape, boxes, img0_shape, ratio_pad=None, padding=True, xywh=False):
    """
    Rescales bounding boxes (in the format of xyxy by default) from the shape of the image they were originally
    specified in (img1_shape) to the shape of a different image (img0_shape).

    Args:
        img1_shape (tuple): The shape of the image that the bounding boxes are for, in the format of (height, width).
        boxes (torch.Tensor): the bounding boxes of the objects in the image, in the format of (x1, y1, x2, y2)
        img0_shape (tuple): the shape of the target image, in the format of (height, width).
        ratio_pad (tuple): a tuple of (ratio, pad) for scaling the boxes. If not provided, the ratio and pad will be
            calculated based on the size difference between the two images.
        padding (bool): If True, assuming the boxes is based on image augmented by yolo style. If False then do regular
            rescaling.
        xywh (bool): The box format is xywh or not, default=False.

    Returns:
        boxes (torch.Tensor): The scaled bounding boxes, in the format of (x1, y1, x2, y2)
    """
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (
            round((img1_shape[1] - img0_shape[1] * gain) / 2 - 0.1),
            round((img1_shape[0] - img0_shape[0] * gain) / 2 - 0.1),
        )  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    if padding:
        boxes[..., 0] -= pad[0]  # x padding
        boxes[..., 1] -= pad[1]  # y padding
        if not xywh:
            boxes[..., 2] -= pad[0]  # x padding
            boxes[..., 3] -= pad[1]  # y padding
    boxes[..., :4] /= gain
    return clip_boxes(boxes, img0_shape)


def clip_boxes(boxes, shape):
    """
    Takes a list of bounding boxes and a shape (height, width) and clips the bounding boxes to the shape.

    Args:
        boxes (torch.Tensor): the bounding boxes to clip
        shape (tuple): the shape of the image

    Returns:
        (torch.Tensor | numpy.ndarray): Clipped boxes
    """
    if isinstance(boxes, torch.Tensor):  # faster individually (WARNING: inplace .clamp_() Apple MPS bug)
        boxes[..., 0] = boxes[..., 0].clamp(0, shape[1])  # x1
        boxes[..., 1] = boxes[..., 1].clamp(0, shape[0])  # y1
        boxes[..., 2] = boxes[..., 2].clamp(0, shape[1])  # x2
        boxes[..., 3] = boxes[..., 3].clamp(0, shape[0])  # y2
    else:  # np.array (faster grouped)
        boxes[..., [0, 2]] = boxes[..., [0, 2]].clip(0, shape[1])  # x1, x2
        boxes[..., [1, 3]] = boxes[..., [1, 3]].clip(0, shape[0])  # y1, y2
    return boxes


def process_mask(protos, masks_in, bboxes, shape, upsample=False):
    """
    Apply masks to bounding boxes using the output of the mask head.

    Args:
        protos (torch.Tensor): A tensor of shape [mask_dim, mask_h, mask_w].
        masks_in (torch.Tensor): A tensor of shape [n, mask_dim], where n is the number of masks after NMS.
        bboxes (torch.Tensor): A tensor of shape [n, 4], where n is the number of masks after NMS.
        shape (tuple): A tuple of integers representing the size of the input image in the format (h, w).
        upsample (bool): A flag to indicate whether to upsample the mask to the original image size. Default is False.

    Returns:
        (torch.Tensor): A binary mask tensor of shape [n, h, w], where n is the number of masks after NMS, and h and w
            are the height and width of the input image. The mask is applied to the bounding boxes.
    """

    c, mh, mw = protos.shape  # CHW
    ih, iw = shape

    masks = (masks_in @ protos.float().view(c, -1)).sigmoid().view(-1, mh, mw)  # CHW
    width_ratio = mw / iw
    height_ratio = mh / ih

    downsampled_bboxes = bboxes.clone()
    downsampled_bboxes[:, 0] *= width_ratio
    downsampled_bboxes[:, 2] *= width_ratio
    downsampled_bboxes[:, 3] *= height_ratio
    downsampled_bboxes[:, 1] *= height_ratio

    masks = crop_mask(masks, downsampled_bboxes)  # CHW
    if upsample:
        masks = F.interpolate(masks[None], shape, mode="bilinear", align_corners=False)[0]  # CHW
    return masks.gt_(0.5)


def crop_mask(masks, boxes):
    """
    It takes a mask and a bounding box, and returns a mask that is cropped to the bounding box.

    Args:
        masks (torch.Tensor): [n, h, w] tensor of masks
        boxes (torch.Tensor): [n, 4] tensor of bbox coordinates in relative point form

    Returns:
        (torch.Tensor): The masks are being cropped to the bounding box.
    """
    _, h, w = masks.shape
    x1, y1, x2, y2 = torch.chunk(boxes[:, :, None], 4, 1)  # x1 shape(n,1,1)
    r = torch.arange(w, device=masks.device, dtype=x1.dtype)[None, None, :]  # rows shape(1,1,w)
    c = torch.arange(h, device=masks.device, dtype=x1.dtype)[None, :, None]  # cols shape(1,h,1)

    return masks * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))


def scale_masks(masks, shape, padding=True):
    """
    Rescale segment masks to shape.

    Args:
        masks (torch.Tensor): (N, C, H, W).
        shape (tuple): Height and width.
        padding (bool): If True, assuming the boxes is based on image augmented by yolo style. If False then do regular
            rescaling.
    """
    mh, mw = masks.shape[2:]
    gain = min(mh / shape[0], mw / shape[1])  # gain  = old / new
    pad = [mw - shape[1] * gain, mh - shape[0] * gain]  # wh padding
    if padding:
        pad[0] /= 2
        pad[1] /= 2
    top, left = (int(pad[1]), int(pad[0])) if padding else (0, 0)  # y, x
    bottom, right = (int(mh - pad[1]), int(mw - pad[0]))
    masks = masks[..., top:bottom, left:right]

    masks = F.interpolate(masks, shape, mode="bilinear", align_corners=False)  # NCHW
    return masks



import logging
from pathlib import Path

from alembic import command
from alembic.config import Config
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.pool import NullPool
from sqlmodel import Session, SQLModel, create_engine

from highlighter.core.config import HighlighterRuntimeConfig

__all__ = [
    "Database",
]


def _set_sqlite_pragma(dbapi_conn, connection_record):
    """Configure SQLite connection for better concurrent access.

    This function is called automatically on each new database connection.
    """
    cursor = dbapi_conn.cursor()
    try:
        cursor.execute("PRAGMA journal_mode=WAL")  # Write-Ahead Logging for concurrency
        cursor.execute("PRAGMA synchronous=NORMAL")  # Faster with WAL mode
        cursor.execute("PRAGMA cache_size=-64000")  # 64MB cache
        cursor.execute("PRAGMA temp_store=MEMORY")  # Use memory for temp tables
    finally:
        cursor.close()


class Database:
    """SQLite-backed metadata store used by the Runtime and Recorder.

    ``Database`` always ensures the schema is on the latest Alembic revision
    before exposing an engine. The pytest suite relies on
    ``RuntimeDatabaseTemplate`` (see ``highlighter.core.database.test_database``)
    to run ``Database`` once at session start, capture the migrated file, and
    cheaply reset the on-disk database before each test. Production code keeps
    calling ``Database`` directly.
    """

    def __init__(self):
        hl_cfg = HighlighterRuntimeConfig.load()
        self.highlighter_path_to_database_file = str(hl_cfg.agent.db_file())
        self._engine = create_engine(
            f"sqlite:///{self.highlighter_path_to_database_file}",
            connect_args={
                "check_same_thread": False,
                "timeout": 30,  # 30-second busy timeout
            },
            poolclass=NullPool,  # Disable pooling for SQLite - connections created/closed per use
        )

        # Configure SQLite for better concurrent access
        event.listens_for(self._engine, "connect")(_set_sqlite_pragma)

        # Run Alembic migrations to ensure schema is up to date
        self._run_migrations()

        # Create any tables that don't exist (for new databases)
        SQLModel.metadata.create_all(self._engine)

    def _run_migrations(self):
        """Run Alembic migrations to upgrade database schema."""
        from alembic.migration import MigrationContext
        from sqlalchemy import inspect

        # Get the directory containing alembic.ini
        alembic_dir = Path(__file__).parent
        alembic_ini_path = alembic_dir / "alembic.ini"

        if not alembic_ini_path.exists():
            # No migrations configured, skip
            return

        # Configure Alembic with absolute paths
        alembic_cfg = Config(str(alembic_ini_path))
        alembic_cfg.set_main_option("script_location", str(alembic_dir / "alembic"))
        alembic_cfg.set_main_option("sqlalchemy.url", f"sqlite:///{self.highlighter_path_to_database_file}")
        alembic_cfg.config_file_name = None  # Prevent alembic/env.py from reconfiguring logging
        self._configure_alembic_logging(alembic_cfg)

        # Check if database has migration history
        with self._engine.begin() as connection:
            context = MigrationContext.configure(connection)
            current_rev = context.get_current_revision()

            # Check if datafile table exists
            inspector = inspect(self._engine)
            tables = inspector.get_table_names()

            if current_rev is None and "datafile" in tables:
                # Database exists but no migration history - stamp it with initial migration
                command.stamp(alembic_cfg, "839d46550586")  # Initial migration revision

        # Run migrations to latest version
        command.upgrade(alembic_cfg, "head")

    def _configure_alembic_logging(self, alembic_cfg: Config) -> None:
        """Route Alembic logs through the Highlighter runtime handlers.

        Alembic's default behaviour is to read logging settings from
        ``alembic.ini`` and emit directly to stderr. That results in noisy test
        output and – more importantly – skips the application's RotatingFile
        handler. By disabling Alembic's logging configuration step and
        synchronising the ``alembic`` logger's level with the Highlighter logger
        we ensure migration messages respect ``HighlighterRuntimeConfig``
        settings and end up in the standard runtime log file.
        """

        # Skip Alembic's logging.config.fileConfig to avoid duplicate handlers
        alembic_cfg.attributes["configure_logger"] = False

        highlighter_logger = logging.getLogger("highlighter")
        # Inherit from root if the highlighter logger hasn't been configured yet
        effective_level = highlighter_logger.getEffectiveLevel()

        alembic_logger = logging.getLogger("alembic")
        alembic_logger.handlers.clear()
        alembic_logger.setLevel(effective_level)
        alembic_logger.propagate = True

    @property
    def engine(self) -> Engine:
        return self._engine

    def get_session(self):
        return Session(self.engine)

    def close(self):
        self._engine.dispose()  # Call this when shutting down your app



import logging
from datetime import datetime
from typing import Dict, Iterator, List, Optional, Union
from uuid import UUID

from highlighter.core.enums import ContentTypeEnum
from highlighter.core.gql_base_model import GQLBaseModel

from .entity import Entity

logger = logging.getLogger(__name__)


class Entities(GQLBaseModel):
    """Entity container

    Enables erganomic management of a set of entities.
    Entities can be looked-up by ID:
        `entity = entities[entity_id]`
    Entities can be added:
        `entities[entity_id] = entity`
        `entities.add(entity)`
        `entities.update(other_entities)`
    Entities can be queried (not yet implemented):
        `specific_entities = entities.where(object_class=object_class_id)`
        `specific_entities = entities.where(has_attribute=attribute_id)`
        `specific_entities = entities.where(has_attribute_value=enum_id)`
    """

    def __init__(self, entities: Optional[Dict[UUID, Entity]] = None):
        super().__init__()
        self._entities = entities or {}

    def add(self, entity: Entity):
        self[entity.id] = entity

    def _clear_old_entities(self, sunset: datetime):
        """Remove stale Entities/Annotations/Observations

        conditions:
          - If Entity.annotations[i].observations[j].occurred_at is too old it is removed
          - If Entity.annotations[i] has no observations and .occurred_at is too old it is removed
          - If Entity.global_observations[i].occurred_at is too old it is removed
          - If an Entity has no .annotations or .global_observations it is removed

        """
        entities_to_remove = []

        for entity_id, entity in self._entities.items():
            # Clean up observations within annotations
            for annotation in list(entity.annotations):
                # Remove old observations from this annotation
                observations_to_remove = [obs for obs in annotation.observations if obs.occurred_at < sunset]
                for obs in observations_to_remove:
                    annotation.observations.remove(obs)

                # Remove annotation if it has no observations and is too old
                if (
                    len(annotation.observations) == 0
                    and annotation.occurred_at
                    and annotation.occurred_at < sunset
                ):
                    entity.annotations.remove(annotation)

            # Clean up global observations
            global_observations_to_remove = [
                obs for obs in entity.global_observations if obs.occurred_at < sunset
            ]
            for obs in global_observations_to_remove:
                entity.global_observations.remove(obs)

            # Mark entity for removal if it has no data left
            if len(entity.annotations) == 0 and len(entity.global_observations) == 0:
                entities_to_remove.append(entity_id)

        # Remove empty entities
        for entity_id in entities_to_remove:
            del self._entities[entity_id]

    def clear(self, sunset: Optional[datetime] = None):
        if sunset:
            self._clear_old_entities(sunset)
        else:
            self._entities.clear()

    def values(self):
        return self._entities.values()

    def items(self):
        return self._entities.items()

    def keys(self):
        return self._entities.keys()

    def __getitem__(self, key: UUID | int):
        if isinstance(key, int):
            return list(self._entities.values())[key]
        return self._entities[key]

    def __delitem__(self, key: UUID):
        return self._entities.__delitem__(key)

    def remove(self, entity: Entity):
        del self[entity.id]

    def __len__(self) -> int:
        return len(self._entities)

    def __iter__(self):
        return iter(list(self._entities.values()))

    def get(self, key: UUID, default: Entity | None = None):
        return self._entities.get(key, default)

    def __setitem__(self, entity_id: UUID, entity: Entity):
        if isinstance(entity, Entity):
            self._entities[entity_id] = entity
        else:
            raise ValueError(f"Expected an Entity, got {type(entity).__qualname__}: {entity}")

    def update(self, *args, **kwargs):
        # Handles both dicts and iterable of pairs
        if args:
            other = args[0]
            if hasattr(other, "items"):
                for k, v in other.items():
                    self[k] = v  # goes through __setitem__
            else:
                for value in other:
                    if isinstance(value, Entity):
                        self.add(value)
                    else:
                        k, v = value
                        self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def __ior__(self, other):
        self.update(other)
        return self

    def __or__(self, other):
        new = type(self)(entities=self._entities.copy())
        new.update(other)
        return new

    def __contains__(self, key):
        return key in self._entities

    def __repr__(self):
        return self._entities.__repr__()

    def to_json_serializable_dict(self):
        return {str(id): entity.to_json() for id, entity in self._entities.items()}

    def to_data_sample(self) -> "DataSample":
        from highlighter.core.data_models.data_sample import DataSample

        if len(self._entities) == 0:
            return DataSample(content=self, content_type=ContentTypeEnum.ENTITIES)
        some_annotations = list(self._entities.values())[0].annotations
        if len(some_annotations) == 0:
            raise ValueError("Cannot convert Entities to DataSample if there are no annotations")
        annotation = some_annotations[0]
        if annotation.data_file_id is None:
            raise ValueError("Cannot convert Entities to DataSample if annotation.data_file_id is None")
        return DataSample(
            content=self,
            content_type=ContentTypeEnum.ENTITIES,
            recorded_at=annotation.occurred_at,
            stream_frame_index=annotation.datum_source.frame_id,
            media_frame_index=annotation.datum_source.frame_id,  # FIXME
        )

    def all_observations(self) -> Iterator["Observation"]:
        for entity in self._entities.values():
            yield from entity.all_observations()

    def to_observations_table(self, stream_id: str, data_sample: "DataSample"):
        """
        Convert Entities to an ObservationsTable.

        Creates one row per annotation (entity + annotation pair). For entities with
        global observations but no annotations, creates one row per entity with a
        placeholder annotation.

        Args:
            stream_id: Optional stream identifier to include in the table

        Returns:
            ObservationsTable instance
        """
        from highlighter.agent.observations_table import ObservationsTable

        rows = []

        for entity in self._entities.values():
            row_data = ObservationsTable.row_data_from_entity(entity, data_sample, stream_id)
            rows.extend(row_data)

        return ObservationsTable.from_row_records(rows)

    def merge(self, others: Union["Entities", Dict], strategy="append"):
        for e in others.values():
            if e.id in self:
                self[e.id].merge(e, strategy=strategy)
            else:
                self.add(e)

    def hydrate_with_data_samples(self, data_samples: List["DataSample"]) -> "Entities":
        for entity in self:
            for annotation in entity.annotations:
                for data_sample in data_samples:
                    # TODO: Use annotation.data_source_id and annotation.occurred_at to resolve data sample
                    if (
                        data_sample.data_file_id == annotation.data_file_id
                        and data_sample.media_frame_index == annotation.datum_source.frame_id
                    ):
                        annotation.data_sample = data_sample
                        break
                else:
                    logger.warning(f"Annotation {annotation.id} has no associated data sample")
        return self



# To Do
# ~~~~~
# - Implement character-per-record option, e.g tty://mode=raw ?
#
# - Implment TextReadTTY parameter {"tty_history": N} for command line history
#   - "N" is maximum command lines kept and "N=0" means no history (default)
#
# - Prompt uses "{}" template to optionally insert "len(tty_command_lines)"

import queue
from threading import Thread

import aiko_services as aiko

__all__ = ["DataSchemeTTY"]

# --------------------------------------------------------------------------- #
# parameters: "data_sources" provides the "tty" scheme URL
# - "data_sources" list should only contain a single entry
#   - "(tty://)"
#
# parameters: "data_targets" provides the "tty" scheme URL
# - "data_targets" list should only contain a single entry
#   - "(tty://)"

class DataSchemeTTY(aiko.DataScheme):
    def create_sources(self,
        stream, data_sources, frame_generator, use_create_frame):

        self.queue = queue.Queue()
        self.terminate = False
        Thread(target=self._run, daemon=True).start()

        rate, _ = self.pipeline_element.get_parameter("rate", default=20.0)
        self.pipeline_element.create_frames(
            stream, self.frame_generator, rate=float(rate))
        return aiko.StreamEvent.OKAY, {}

    def create_targets(self, stream, data_targets):
        return aiko.StreamEvent.OKAY, {}

    def destroy_sources(self, stream):
        self.terminate = True

    def frame_generator(self, stream, frame_id):
        if self.queue.qsize():
            record = self.queue.get()
            if record is not None:
                return aiko.StreamEvent.OKAY, {"records": [record]}
            else:
                diagnostic = "All frames generated"
                return aiko.StreamEvent.STOP, {"diagnostic": diagnostic}
        else:
            return aiko.StreamEvent.NO_FRAME, {}

    def _run(self):
        tty_prompt, _ = self.pipeline_element.get_parameter(
                            "tty_prompt", default="> ")
        print(f"000:{tty_prompt}", end="")
        while not self.terminate:
            try:
                self.queue.put(input())
            except EOFError:
                self.queue.put(None)

aiko.DataScheme.add_data_scheme("tty", DataSchemeTTY)

# --------------------------------------------------------------------------- #
"""
import threading
import queue
import sys
import termios
import tty

def non_blocking_keyboard_input():
    input_queue = queue.Queue()

    def input_reader():
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            while True:
                user_input = sys.stdin.read(1)  # Read one character
                input_queue.put(user_input)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    input_thread = threading.Thread(target=input_reader, daemon=True)
    input_thread.start()

    while True:
        try:
            user_input = input_queue.get(timeout=0.1)
            yield user_input
        except queue.Empty:
            continue
"""
# --------------------------------------------------------------------------- #
